@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2010-2019 Marcel Martin <marcel.martin@scilifelab.se>
+# Copyright (c) 2010-2020 Marcel Martin <marcel.martin@scilifelab.se>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -23,7 +23,7 @@
 """
 cutadapt version {version}
 
-Copyright (C) 2010-2019 Marcel Martin <marcel.martin@scilifelab.se>
+Copyright (C) 2010-2020 Marcel Martin <marcel.martin@scilifelab.se>
 
 cutadapt removes adapter sequences from high-throughput sequencing reads.
 
@@ -57,7 +57,7 @@ import sys
 import time
 import logging
 import platform
-from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union
+from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union, Type
 from argparse import ArgumentParser, SUPPRESS, HelpFormatter
 
 import dnaio
@@ -65,13 +65,13 @@ import dnaio
 from cutadapt import __version__
 from cutadapt.adapters import warn_duplicate_adapters, Adapter
 from cutadapt.parser import AdapterParser
-from cutadapt.modifiers import (Modifier, LengthTagModifier, SuffixRemover, PrefixSuffixAdder,
+from cutadapt.modifiers import (SingleEndModifier, LengthTagModifier, SuffixRemover, PrefixSuffixAdder,
     ZeroCapper, QualityTrimmer, UnconditionalCutter, NEndTrimmer, AdapterCutter,
     PairedAdapterCutterError, PairedAdapterCutter, NextseqQualityTrimmer, Shortener,
     ReverseComplementer)
 from cutadapt.report import full_report, minimal_report
 from cutadapt.pipeline import (Pipeline, SingleEndPipeline, PairedEndPipeline, InputFiles,
-    OutputFiles, SerialPipelineRunner, ParallelPipelineRunner)
+    OutputFiles, PipelineRunner, SerialPipelineRunner, ParallelPipelineRunner)
 from cutadapt.utils import available_cpu_count, Progress, DummyProgress, FileOpener
 from cutadapt.log import setup_logging, REPORT
 
@@ -116,8 +116,9 @@ def get_argument_parser() -> ArgumentParser:
     group.add_argument("-h", "--help", action="help", help="Show this help message and exit")
     group.add_argument("--version", action="version", help="Show version number and exit",
         version=__version__)
-    group.add_argument("--debug", action='store_true', default=False,
-        help="Print debugging information.")
+    group.add_argument("--debug", nargs="?", const=True, default=False,
+        choices=("trace",),
+        help="Print debug log. 'trace' prints also DP matrices")
     group.add_argument("--profile", action="store_true", default=False, help=SUPPRESS)
     group.add_argument('-j', '--cores', type=int, default=1,
         help='Number of CPU cores to use. Use 0 to auto-detect. Default: %(default)s')
@@ -248,6 +249,10 @@ def get_argument_parser() -> ArgumentParser:
     group.add_argument("--max-n", type=float, default=None, metavar="COUNT",
         help="Discard reads with more than COUNT 'N' bases. If COUNT is a number "
              "between 0 and 1, it is interpreted as a fraction of the read length.")
+    group.add_argument("--max-expected-errors", "--max-ee", type=float, default=None,
+        metavar="ERRORS",
+        help="Discard reads whose expected number of errors (computed "
+            "from quality values) exceeds ERRORS.")
     group.add_argument("--discard-trimmed", "--discard", action='store_true', default=False,
         help="Discard reads that contain an adapter. Use also -O to avoid "
             "discarding too many randomly matching reads.")
@@ -316,7 +321,7 @@ def get_argument_parser() -> ArgumentParser:
             "filtering criterion in order for the pair to be filtered. "
             "Default: any")
     group.add_argument("--interleaved", action='store_true', default=False,
-        help="Read and write interleaved paired-end reads.")
+        help="Read and/or write interleaved paired-end reads.")
     group.add_argument("--untrimmed-paired-output", metavar="FILE",
         help="Write second read in a pair to this FILE when no adapter "
             "was found. Use with --untrimmed-output. Default: output "
@@ -388,7 +393,7 @@ def parse_lengths(s: str) -> Tuple[Optional[int], ...]:
 
 
 def open_output_files(
-    args, default_outfile, interleaved: bool, file_opener: FileOpener
+    args, default_outfile, file_opener: FileOpener
 ) -> OutputFiles:
     """
     Return an OutputFiles instance. If demultiplex is True, the untrimmed, untrimmed2, out and out2
@@ -467,7 +472,6 @@ def open_output_files(
         out=out,
         out2=out2,
         demultiplex=bool(demultiplex_mode),
-        interleaved=interleaved,
         force_fasta=args.fasta,
     )
 
@@ -511,22 +515,11 @@ def determine_paired(args) -> bool:
         or args.adapters2
         or args.cut2
         or args.pair_filter
+        or args.untrimmed_paired_output
         or args.too_short_paired_output
         or args.too_long_paired_output)
 
 
-def determine_interleaved(args) -> Tuple[bool, bool]:
-    is_interleaved_input = False
-    is_interleaved_output = False
-    if args.interleaved:
-        is_interleaved_input = len(args.inputs) == 1
-        is_interleaved_output = not args.paired_output
-        if not is_interleaved_input and not is_interleaved_output:
-            raise CommandLineError("When --interleaved is used, you cannot provide both two "
-                "input files and two output files")
-    return is_interleaved_input, is_interleaved_output
-
-
 def setup_input_files(
     inputs: Sequence[str], paired: bool, interleaved: bool
 ) -> Tuple[str, Optional[str]]:
@@ -541,7 +534,7 @@ def setup_input_files(
             "You provided {} input file names, but either one or two are expected. ".format(
                 len(inputs))
             + "The file names were:\n - "
-            + "\n - ".join("{!r}".format(p) for p in inputs)
+            + "\n - ".join("'{}'".format(p) for p in inputs)
             + "\nHint: If your path contains spaces, you need to enclose it in quotes")
     input_filename = inputs[0]
     if paired and not interleaved:
@@ -564,7 +557,7 @@ def setup_input_files(
     return input_filename, input_paired_filename
 
 
-def check_arguments(args, paired: bool, is_interleaved_output: bool) -> None:
+def check_arguments(args, paired: bool) -> None:
     if not paired:
         if args.untrimmed_paired_output:
             raise CommandLineError("Option --untrimmed-paired-output can only be used when "
@@ -574,24 +567,23 @@ def check_arguments(args, paired: bool, is_interleaved_output: bool) -> None:
             raise CommandLineError("Option --pair-adapters can only be used when trimming "
                 "paired-end reads")
 
-    if paired:
-        if not is_interleaved_output:
-            if not args.paired_output:
-                raise CommandLineError("When a paired-end trimming option such as -A/-G/-B/-U, "
-                    "is used, a second output file needs to be specified via -p (--paired-output).")
-            if not args.output:
-                raise CommandLineError("When you use -p or --paired-output, you must also "
-                    "use the -o option.")
-
-        if bool(args.untrimmed_output) != bool(args.untrimmed_paired_output):
-            raise CommandLineError("When trimming paired-end reads, you must use either none "
-                "or both of the --untrimmed-output/--untrimmed-paired-output options.")
-        if args.too_short_output and not args.too_short_paired_output:
-            raise CommandLineError("When using --too-short-output with paired-end "
-                "reads, you also need to use --too-short-paired-output")
-        if args.too_long_output and not args.too_long_paired_output:
-            raise CommandLineError("When using --too-long-output with paired-end "
-                "reads, you also need to use --too-long-paired-output")
+    if paired and not args.interleaved:
+        if not args.paired_output:
+            raise CommandLineError("When a paired-end trimming option such as -A/-G/-B/-U, "
+                "is used, a second output file needs to be specified via -p (--paired-output).")
+        if not args.output:
+            raise CommandLineError("When you use -p or --paired-output, you must also "
+                "use the -o option.")
+        for out, paired_out, argname in [
+            (args.untrimmed_output, args.untrimmed_paired_output, "untrimmed"),
+            (args.too_short_output, args.too_short_paired_output, "too-short"),
+            (args.too_long_output, args.too_long_paired_output, "too-long"),
+        ]:
+            if bool(out) != bool(paired_out):
+                raise CommandLineError(
+                    "When trimming paired-end data, you must use either none or both of the"
+                    " --{name}-output/--{name}-paired-output options.".format(name=argname)
+                )
 
     if args.format is not None:
         logger.warning("Option --format is deprecated and ignored because the input file format is "
@@ -672,6 +664,7 @@ def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
                 lengths = (lengths[0], lengths[0])
             setattr(pipeline, attr, lengths)
     pipeline.max_n = args.max_n
+    pipeline.max_expected_errors = args.max_expected_errors
     pipeline.discard_casava = args.discard_casava
     pipeline.discard_trimmed = args.discard_trimmed
     pipeline.discard_untrimmed = args.discard_untrimmed
@@ -694,7 +687,7 @@ def adapters_from_args(args) -> Tuple[List[Adapter], List[Adapter]]:
         raise CommandLineError(e)
     warn_duplicate_adapters(adapters)
     warn_duplicate_adapters(adapters2)
-    if args.debug:
+    if args.debug == "trace":
         for adapter in adapters + adapters2:
             adapter.enable_debug()
     return adapters, adapters2
@@ -764,7 +757,7 @@ def add_adapter_cutter(
             pipeline.add(modifier)
 
 
-def modifiers_applying_to_both_ends_if_paired(args) -> Iterator[Modifier]:
+def modifiers_applying_to_both_ends_if_paired(args) -> Iterator[SingleEndModifier]:
     if args.length is not None:
         yield Shortener(args.length)
     if args.trim_n:
@@ -800,7 +793,7 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     parser = get_argument_parser()
     if cmdlineargs is None:
         cmdlineargs = sys.argv[1:]
-    args = parser.parse_args(args=cmdlineargs)
+    args, leftover_args = parser.parse_known_args(args=cmdlineargs)
     # log to stderr if results are to be sent to stdout
     log_to_stdout = args.output is not None and args.output != "-" and args.paired_output != "-"
     # Setup logging only if there are not already any handlers (can happen when
@@ -808,12 +801,7 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     if not logging.root.handlers:
         setup_logging(logger, stdout=log_to_stdout,
             quiet=args.quiet, minimal=args.report == 'minimal', debug=args.debug)
-    if args.profile:
-        import cProfile
-        profiler = cProfile.Profile()
-        profiler.enable()
-    else:
-        profiler = None
+    profiler = setup_profiler_if_requested(args.profile)
 
     if args.quiet and args.report:
         parser.error("Options --quiet and --report cannot be used at the same time")
@@ -830,53 +818,41 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
 
     # Print the header now because some of the functions below create logging output
     log_header(cmdlineargs)
+    if leftover_args:
+        warn_if_en_dashes(cmdlineargs)
+        parser.error("unrecognized arguments: " + " ".join(leftover_args))
+
     if args.cores < 0:
         parser.error('Value for --cores cannot be negative')
 
     cores = available_cpu_count() if args.cores == 0 else args.cores
     file_opener = FileOpener(
         compression_level=args.compression_level, threads=0 if cores == 1 else None)
-    try:
-        is_interleaved_input, is_interleaved_output = determine_interleaved(args)
-        input_filename, input_paired_filename = setup_input_files(args.inputs,
-            paired, is_interleaved_input)
-        check_arguments(args, paired, is_interleaved_output)
-        pipeline = pipeline_from_parsed_args(args, paired, file_opener)
-        outfiles = open_output_files(args, default_outfile, is_interleaved_output, file_opener)
-    except CommandLineError as e:
-        parser.error(str(e))
-        return  # avoid IDE warnings below
-
-    if cores > 1:
-        if ParallelPipelineRunner.can_output_to(outfiles):
-            runner_class = ParallelPipelineRunner
-            runner_kwargs = dict(n_workers=cores, buffer_size=args.buffer_size)
-        else:
-            parser.error("Running in parallel is currently not supported "
-                "when using --format or when demultiplexing.\n"
-                "Omit --cores/-j to continue.")
-            return  # avoid IDE warnings below
-    else:
-        runner_class = SerialPipelineRunner
-        runner_kwargs = dict()
-    infiles = InputFiles(input_filename, file2=input_paired_filename,
-            interleaved=is_interleaved_input)
     if sys.stderr.isatty() and not args.quiet:
         progress = Progress()
     else:
         progress = DummyProgress()
+
     try:
-        runner = runner_class(pipeline, infiles, outfiles, progress, **runner_kwargs)
-    except (dnaio.UnknownFileFormat, IOError) as e:
-        parser.error(e)
+        is_interleaved_input = args.interleaved and len(args.inputs) == 1
+        input_filename, input_paired_filename = setup_input_files(args.inputs,
+            paired, is_interleaved_input)
+        check_arguments(args, paired)
+        pipeline = pipeline_from_parsed_args(args, paired, file_opener)
+        outfiles = open_output_files(args, default_outfile, file_opener)
+        infiles = InputFiles(input_filename, file2=input_paired_filename,
+            interleaved=is_interleaved_input)
+        runner = setup_runner(pipeline, infiles, outfiles, progress, cores, args.buffer_size)
+    except CommandLineError as e:
+        parser.error(str(e))
         return  # avoid IDE warnings below
 
     logger.info("Processing reads on %d core%s in %s mode ...",
         cores, 's' if cores > 1 else '',
         {False: 'single-end', True: 'paired-end'}[pipeline.paired])
     try:
-        stats = runner.run()
-        runner.close()
+        with runner as r:
+            stats = r.run()
     except KeyboardInterrupt:
         print("Interrupted", file=sys.stderr)
         sys.exit(130)
@@ -897,5 +873,46 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
         pstats.Stats(profiler).sort_stats('time').print_stats(20)
 
 
+def setup_runner(pipeline: Pipeline, infiles, outfiles, progress, cores, buffer_size):
+    if cores > 1:
+        if ParallelPipelineRunner.can_output_to(outfiles):
+            runner_class = ParallelPipelineRunner  # type: Type[PipelineRunner]
+            runner_kwargs = dict(n_workers=cores, buffer_size=buffer_size)
+        else:
+            raise CommandLineError("Running in parallel is currently not supported "
+                 "when using --format or when demultiplexing.\n"
+                 "Omit --cores/-j to continue.")
+            # return  # avoid IDE warnings below
+    else:
+        runner_class = SerialPipelineRunner
+        runner_kwargs = dict()
+    try:
+        runner = runner_class(pipeline, infiles, outfiles, progress, **runner_kwargs)
+    except (dnaio.UnknownFileFormat, OSError) as e:
+        raise CommandLineError(e)
+
+    return runner
+
+
+def setup_profiler_if_requested(requested):
+    if requested:
+        import cProfile
+        profiler = cProfile.Profile()
+        profiler.enable()
+    else:
+        profiler = None
+    return profiler
+
+
+def warn_if_en_dashes(args):
+    for arg in args:
+        if arg.startswith("–"):
+            logger.warning(
+                "The first character in argument '%s' is '–' (an en-dash, Unicode U+2013)"
+                " and will therefore be interpreted as a file name. If you wanted to"
+                " provide an option, use a regular hyphen '-'.", arg
+            )
+
+
 if __name__ == '__main__':
     main()
