@@ -9,19 +9,32 @@ from abc import ABC, abstractmethod
 from collections import OrderedDict
 
 from .qualtrim import quality_trim_index, nextseq_trim_index
-from .adapters import Where, MultiAdapter, Match, remainder
+from .adapters import MultiPrefixAdapter, MultiSuffixAdapter, Match, remainder
 from .utils import reverse_complemented_sequence
 
 
-class Modifier(ABC):
+class ModificationInfo:
+    """
+    An object of this class is created for each read that passes through the pipeline.
+    Any information (except the read itself) that needs to be passed from one modifier
+    to one later in the pipeline or from one modifier to the filters is recorded here.
+    """
+    __slots__ = ["matches", "original_read"]
+
+    def __init__(self, read):
+        self.matches = []  # type: List[Match]
+        self.original_read = read
+
+
+class SingleEndModifier(ABC):
     @abstractmethod
-    def __call__(self, read, matches: List[Match]):
+    def __call__(self, read, info: ModificationInfo):
         pass
 
 
 class PairedModifier(ABC):
     @abstractmethod
-    def __call__(self, read1, read2, matches1, matches2):
+    def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
         pass
 
 
@@ -31,24 +44,26 @@ class PairedModifierWrapper(PairedModifier):
     """
     paired = True
 
-    def __init__(self, modifier1: Optional[Modifier], modifier2: Optional[Modifier]):
+    def __init__(self, modifier1: Optional[SingleEndModifier], modifier2: Optional[SingleEndModifier]):
         """Set one of the modifiers to None to work on R1 or R2 only"""
         self._modifier1 = modifier1
         self._modifier2 = modifier2
+        if self._modifier1 is None and self._modifier2 is None:
+            raise ValueError("Not both modifiers may be None")
 
     def __repr__(self):
         return 'PairedModifier(modifier1={!r}, modifier2={!r})'.format(
             self._modifier1, self._modifier2)
 
-    def __call__(self, read1, read2, matches1, matches2):
+    def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
         if self._modifier1 is None:
-            return read1, self._modifier2(read2, matches2)
+            return read1, self._modifier2(read2, info2)  # type: ignore
         if self._modifier2 is None:
-            return self._modifier1(read1, matches1), read2
-        return self._modifier1(read1, matches1), self._modifier2(read2, matches2)
+            return self._modifier1(read1, info1), read2
+        return self._modifier1(read1, info1), self._modifier2(read2, info2)
 
 
-class AdapterCutter(Modifier):
+class AdapterCutter(SingleEndModifier):
     """
     Repeatedly find one of multiple adapters in reads.
     The number of times the search is repeated is specified by the
@@ -71,11 +86,14 @@ class AdapterCutter(Modifier):
         # the adapters when we don’t need to
         if len(prefix) > 1 or len(suffix) > 1:
             adapters = other
-            for affix in (prefix, suffix):
-                if len(affix) > 1:
-                    adapters.append(MultiAdapter(affix))
-                else:
-                    adapters.extend(affix)
+            if len(prefix) > 1:
+                adapters.append(MultiPrefixAdapter(prefix))
+            else:
+                adapters.extend(prefix)
+            if len(suffix) > 1:
+                adapters.append(MultiSuffixAdapter(suffix))
+            else:
+                adapters.extend(suffix)
         self.adapters = adapters
 
     def __repr__(self):
@@ -93,12 +111,10 @@ class AdapterCutter(Modifier):
         """
         prefix, suffix, other = [], [], []
         for a in adapters:
-            if MultiAdapter.is_acceptable(a):
-                if a.where == Where.PREFIX:
-                    lst = prefix
-                else:
-                    assert a.where == Where.SUFFIX
-                    lst = suffix
+            if MultiPrefixAdapter.is_acceptable(a):
+                lst = prefix
+            elif MultiSuffixAdapter.is_acceptable(a):
+                lst = suffix
             else:
                 lst = other
             lst.append(a)
@@ -113,7 +129,7 @@ class AdapterCutter(Modifier):
         """
         best_match = None
         for adapter in adapters:
-            match = adapter.match_to(read)
+            match = adapter.match_to(read.sequence)
             if match is None:
                 continue
 
@@ -125,36 +141,33 @@ class AdapterCutter(Modifier):
         return best_match
 
     @staticmethod
-    def masked_read(read, trimmed_read, matches: Sequence[Match]):
+    def masked_read(read, matches: Sequence[Match]):
         start, stop = remainder(matches)
-        # TODO modification in place
-        trimmed_read.sequence = (
+        result = read[:]
+        result.sequence = (
             'N' * start
             + read.sequence[start:stop]
             + 'N' * (len(read) - stop))
-        trimmed_read.qualities = read.qualities
-        return trimmed_read
+        return result
 
     @staticmethod
-    def lowercased_read(read, trimmed_read, matches: Sequence[Match]):
+    def lowercased_read(read, matches: Sequence[Match]):
         start, stop = remainder(matches)
-        read_sequence = read.sequence
-        # TODO modification in place
-        trimmed_read.sequence = (
-            read_sequence[:start].lower()
-            + read_sequence[start:stop].upper()
-            + read_sequence[stop:].lower()
+        result = read[:]
+        result.sequence = (
+            read.sequence[:start].lower()
+            + read.sequence[start:stop].upper()
+            + read.sequence[stop:].lower()
         )
-        trimmed_read.qualities = read.qualities
-        return trimmed_read
+        return result
 
-    def __call__(self, read, inmatches: List[Match]):
+    def __call__(self, read, info: ModificationInfo):
         trimmed_read, matches = self.match_and_trim(read)
         if matches:
             self.with_adapters += 1
             for match in matches:
                 match.update_statistics(self.adapter_statistics[match.adapter])
-        inmatches.extend(matches)
+        info.matches.extend(matches)  # TODO extend or overwrite?
         return trimmed_read
 
     def match_and_trim(self, read):
@@ -179,7 +192,7 @@ class AdapterCutter(Modifier):
                 # if nothing found, attempt no further rounds
                 break
             matches.append(match)
-            trimmed_read = match.trimmed()
+            trimmed_read = match.trimmed(trimmed_read)
 
         if not matches:
             return trimmed_read, []
@@ -188,9 +201,9 @@ class AdapterCutter(Modifier):
             # read is already trimmed, nothing to do
             pass
         elif self.action == 'mask':
-            trimmed_read = self.masked_read(read, trimmed_read, matches)
+            trimmed_read = self.masked_read(read, matches)
         elif self.action == 'lowercase':
-            trimmed_read = self.lowercased_read(read, trimmed_read, matches)
+            trimmed_read = self.lowercased_read(read, matches)
             assert len(trimmed_read.sequence) == len(read)
         elif self.action is None:  # --no-trim
             trimmed_read = read[:]
@@ -198,7 +211,7 @@ class AdapterCutter(Modifier):
         return trimmed_read, matches
 
 
-class ReverseComplementer(Modifier):
+class ReverseComplementer(SingleEndModifier):
     """Trim adapters from a read and its reverse complement"""
 
     def __init__(self, adapter_cutter: AdapterCutter, rc_suffix: Optional[str] = " rc"):
@@ -209,7 +222,7 @@ class ReverseComplementer(Modifier):
         self.reverse_complemented = 0
         self._suffix = rc_suffix
 
-    def __call__(self, read, inmatches: List[Match]):
+    def __call__(self, read, info: ModificationInfo):
         reverse_read = reverse_complemented_sequence(read)
 
         forward_trimmed_read, forward_matches = self.adapter_cutter.match_and_trim(read)
@@ -234,7 +247,7 @@ class ReverseComplementer(Modifier):
                 stats = self.adapter_cutter.adapter_statistics[match.adapter]
                 match.update_statistics(stats)
                 stats.reverse_complemented += bool(use_reverse_complement)
-            inmatches.extend(matches)
+            info.matches.extend(matches)  # TODO extend or overwrite?
         return trimmed_read
 
 
@@ -255,7 +268,7 @@ class PairedAdapterCutter(PairedModifier):
         Both lists must have the same, non-zero length.
          read pair is trimmed if adapters1[i] is found in R1 and adapters2[i] in R2.
 
-        action -- What to do with a found adapter: None, 'trim', or 'mask'
+        action -- What to do with a found adapter: None, 'trim', 'lowercase' or 'mask'
         """
         super().__init__()
         if len(adapters1) != len(adapters2):
@@ -277,7 +290,7 @@ class PairedAdapterCutter(PairedModifier):
         return 'PairedAdapterCutter(adapters1={!r}, adapters2={!r})'.format(
             self._adapters1, self._adapters2)
 
-    def __call__(self, read1, read2, matches1, matches2):
+    def __call__(self, read1, read2, info1, info2):
         """
         """
         match1 = AdapterCutter.best_match(self._adapters1, read1)
@@ -285,7 +298,7 @@ class PairedAdapterCutter(PairedModifier):
             return read1, read2
         adapter1 = match1.adapter
         adapter2 = self._adapters2[self._adapter_indices[adapter1]]
-        match2 = adapter2.match_to(read2)
+        match2 = adapter2.match_to(read2.sequence)
         if match2 is None:
             return read1, read2
 
@@ -296,26 +309,26 @@ class PairedAdapterCutter(PairedModifier):
             if self.action == 'lowercase':
                 trimmed_read.sequence = trimmed_read.sequence.upper()
 
-            trimmed_read = match.trimmed()
+            trimmed_read = match.trimmed(trimmed_read)
             match.update_statistics(self.adapter_statistics[i][match.adapter])
 
             if self.action == 'trim':
                 # read is already trimmed, nothing to do
                 pass
             elif self.action == 'mask':
-                trimmed_read = AdapterCutter.masked_read(trimmed_read, [match])
+                trimmed_read = AdapterCutter.masked_read(read, [match])
             elif self.action == 'lowercase':
-                trimmed_read = AdapterCutter.lowercased_read(trimmed_read, [match])
+                trimmed_read = AdapterCutter.lowercased_read(read, [match])
                 assert len(trimmed_read.sequence) == len(read)
             elif self.action is None:  # --no-trim
                 trimmed_read = read[:]
             result.append(trimmed_read)
-        matches1.append(match1)
-        matches2.append(match2)
+        info1.matches.append(match1)
+        info2.matches.append(match2)
         return result
 
 
-class UnconditionalCutter(Modifier):
+class UnconditionalCutter(SingleEndModifier):
     """
     A modifier that unconditionally removes the first n or the last n bases from a read.
 
@@ -325,14 +338,14 @@ class UnconditionalCutter(Modifier):
     def __init__(self, length: int):
         self.length = length
 
-    def __call__(self, read, matches: List[Match]):
+    def __call__(self, read, info: ModificationInfo):
         if self.length > 0:
             return read[self.length:]
         elif self.length < 0:
             return read[:self.length]
 
 
-class LengthTagModifier(Modifier):
+class LengthTagModifier(SingleEndModifier):
     """
     Replace "length=..." strings in read names.
     """
@@ -340,28 +353,28 @@ class LengthTagModifier(Modifier):
         self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
         self.length_tag = length_tag
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.find(self.length_tag) >= 0:
             read.name = self.regex.sub(self.length_tag + str(len(read.sequence)), read.name)
         return read
 
 
-class SuffixRemover(Modifier):
+class SuffixRemover(SingleEndModifier):
     """
     Remove a given suffix from read names.
     """
     def __init__(self, suffix):
         self.suffix = suffix
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.endswith(self.suffix):
             read.name = read.name[:-len(self.suffix)]
         return read
 
 
-class PrefixSuffixAdder(Modifier):
+class PrefixSuffixAdder(SingleEndModifier):
     """
     Add a suffix and a prefix to read names
     """
@@ -369,15 +382,15 @@ class PrefixSuffixAdder(Modifier):
         self.prefix = prefix
         self.suffix = suffix
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info):
         read = read[:]
-        adapter_name = matches[-1].adapter.name if matches else 'no_adapter'
+        adapter_name = info.matches[-1].adapter.name if info.matches else 'no_adapter'
         read.name = self.prefix.replace('{name}', adapter_name) + read.name + \
             self.suffix.replace('{name}', adapter_name)
         return read
 
 
-class ZeroCapper(Modifier):
+class ZeroCapper(SingleEndModifier):
     """
     Change negative quality values of a read to zero
     """
@@ -385,38 +398,38 @@ class ZeroCapper(Modifier):
         qb = quality_base
         self.zero_cap_trans = str.maketrans(''.join(map(chr, range(qb))), chr(qb) * qb)
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         read = read[:]
         read.qualities = read.qualities.translate(self.zero_cap_trans)
         return read
 
 
-class NextseqQualityTrimmer(Modifier):
+class NextseqQualityTrimmer(SingleEndModifier):
     def __init__(self, cutoff, base):
         self.cutoff = cutoff
         self.base = base
         self.trimmed_bases = 0
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         stop = nextseq_trim_index(read, self.cutoff, self.base)
         self.trimmed_bases += len(read) - stop
         return read[:stop]
 
 
-class QualityTrimmer(Modifier):
+class QualityTrimmer(SingleEndModifier):
     def __init__(self, cutoff_front, cutoff_back, base):
         self.cutoff_front = cutoff_front
         self.cutoff_back = cutoff_back
         self.base = base
         self.trimmed_bases = 0
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
         self.trimmed_bases += len(read) - (stop - start)
         return read[start:stop]
 
 
-class Shortener(Modifier):
+class Shortener(SingleEndModifier):
     """Unconditionally shorten a read to the given length
 
     If the length is positive, the bases are removed from the end of the read.
@@ -425,20 +438,20 @@ class Shortener(Modifier):
     def __init__(self, length):
         self.length = length
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         if self.length >= 0:
             return read[:self.length]
         else:
             return read[self.length:]
 
 
-class NEndTrimmer(Modifier):
+class NEndTrimmer(SingleEndModifier):
     """Trims Ns from the 3' and 5' end of reads"""
     def __init__(self):
         self.start_trim = re.compile(r'^N+')
         self.end_trim = re.compile(r'N+$')
 
-    def __call__(self, read, matches):
+    def __call__(self, read, info: ModificationInfo):
         sequence = read.sequence
         start_cut = self.start_trim.match(sequence)
         end_cut = self.end_trim.search(sequence)
