@@ -4,7 +4,7 @@ import sys
 import copy
 import logging
 import functools
-from typing import List, IO, Optional, BinaryIO, TextIO, Any, Tuple, Union
+from typing import List, IO, Optional, BinaryIO, TextIO, Any, Tuple
 from abc import ABC, abstractmethod
 from multiprocessing import Process, Pipe, Queue
 from pathlib import Path
@@ -16,10 +16,11 @@ from xopen import xopen
 import dnaio
 
 from .utils import Progress, FileOpener
-from .modifiers import Modifier, PairedModifier, PairedModifierWrapper
+from .modifiers import SingleEndModifier, PairedModifier, PairedModifierWrapper, ModificationInfo
 from .report import Statistics
 from .filters import (Redirector, PairedRedirector, NoFilter, PairedNoFilter, InfoFileWriter,
     RestFileWriter, WildcardFileWriter, TooShortReadFilter, TooLongReadFilter, NContentFilter,
+    MaximumExpectedErrorsFilter,
     CasavaFilter, DiscardTrimmedFilter, DiscardUntrimmedFilter, Demultiplexer,
     PairedDemultiplexer, CombinatorialDemultiplexer)
 
@@ -32,6 +33,10 @@ class InputFiles:
         self.file2 = file2
         self.interleaved = interleaved
 
+    def open(self):
+        return dnaio.open(self.file1, file2=self.file2,
+            interleaved=self.interleaved, mode="r")
+
 
 class OutputFiles:
     """
@@ -39,11 +44,8 @@ class OutputFiles:
     untrimmed, untrimmed2, out and out2 are file names or templates
     as required by the used demultiplexer ('{name}' etc.).
 
-    If interleaved is True, then out is written interleaved.
-
     Files may also be None.
     """
-    # TODO interleaving for the other file pairs (too_short, too_long, untrimmed)?
     def __init__(
         self,
         out: Optional[BinaryIO] = None,
@@ -58,7 +60,6 @@ class OutputFiles:
         rest: Optional[BinaryIO] = None,
         wildcard: Optional[BinaryIO] = None,
         demultiplex: bool = False,
-        interleaved: bool = False,
         force_fasta: Optional[bool] = None,
     ):
         self.out = out
@@ -73,7 +74,6 @@ class OutputFiles:
         self.rest = rest
         self.wildcard = wildcard
         self.demultiplex = demultiplex
-        self.interleaved = interleaved
         self.force_fasta = force_fasta
 
     def __iter__(self):
@@ -98,10 +98,8 @@ class Pipeline(ABC):
 
     def __init__(self, file_opener: FileOpener):
         self._close_files = []  # type: List[IO]
-        self._reader = None
+        self._reader = None  # type: Any
         self._filters = []  # type: List[Any]
-        # TODO type should be Union[List[Modifier], List[PairedModifier]]
-        self._modifiers = []  # type: List[Union[Modifier, PairedModifier]]
         self._outfiles = None  # type: Optional[OutputFiles]
         self._demultiplexer = None
         self._textiowrappers = []  # type: List[TextIO]
@@ -110,33 +108,24 @@ class Pipeline(ABC):
         self._minimum_length = None
         self._maximum_length = None
         self.max_n = None
+        self.max_expected_errors = None
         self.discard_casava = False
         self.discard_trimmed = False
         self.discard_untrimmed = False
         self.file_opener = file_opener
 
     def connect_io(self, infiles: InputFiles, outfiles: OutputFiles):
-        self._reader = dnaio.open(infiles.file1, file2=infiles.file2,
-            interleaved=infiles.interleaved, mode='r')
+        self._reader = infiles.open()
         self._set_output(outfiles)
 
+    @abstractmethod
     def _open_writer(
         self,
         file: BinaryIO,
         file2: Optional[BinaryIO] = None,
         force_fasta: Optional[bool] = None,
-        **kwargs
     ):
-        # TODO backwards-incompatible change (?) would be to use outfiles.interleaved
-        # for all outputs
-        if force_fasta:
-            kwargs['fileformat'] = 'fasta'
-        # file and file2 must already be file-like objects because we don’t want to
-        # take care of threads and compression levels here.
-        for f in (file, file2):
-            assert not isinstance(f, (str, bytes, Path))
-        return dnaio.open(file, file2=file2, mode='w', qualities=self.uses_qualities,
-            **kwargs)
+        pass
 
     def _set_output(self, outfiles: OutputFiles):
         self._filters = []
@@ -172,6 +161,13 @@ class Pipeline(ABC):
             f1 = f2 = NContentFilter(self.max_n)
             self._filters.append(filter_wrapper(None, f1, f2))
 
+        if self.max_expected_errors is not None:
+            if not self._reader.delivers_qualities:
+                logger.warning("Ignoring option --max-ee as input does not contain quality values")
+            else:
+                f1 = f2 = MaximumExpectedErrorsFilter(self.max_expected_errors)
+                self._filters.append(filter_wrapper(None, f1, f2))
+
         if self.discard_casava:
             f1 = f2 = CasavaFilter()
             self._filters.append(filter_wrapper(None, f1, f2))
@@ -211,6 +207,7 @@ class Pipeline(ABC):
                 f.flush()
 
     def close(self) -> None:
+        self._reader.close()
         for f in self._textiowrappers:
             f.close()  # This also closes the underlying files; a second close occurs below
         assert self._outfiles is not None
@@ -255,31 +252,41 @@ class SingleEndPipeline(Pipeline):
 
     def __init__(self, file_opener: FileOpener):
         super().__init__(file_opener)
-        self._modifiers = []
+        self._modifiers = []  # type: List[SingleEndModifier]
 
-    def add(self, modifier: Modifier):
+    def add(self, modifier: SingleEndModifier):
         if modifier is None:
             raise ValueError("Modifier must not be None")
         self._modifiers.append(modifier)
 
     def process_reads(self, progress: Progress = None):
         """Run the pipeline. Return statistics"""
-        n = 0  # no. of processed reads  # TODO turn into attribute
+        n = 0  # no. of processed reads
         total_bp = 0
-        assert self._reader is not None
         for read in self._reader:
             n += 1
             if n % 10000 == 0 and progress:
                 progress.update(n)
-            total_bp += len(read.sequence)
-            matches = []
+            total_bp += len(read)
+            info = ModificationInfo(read)
             for modifier in self._modifiers:
-                read = modifier(read, matches)
+                read = modifier(read, info)
             for filter_ in self._filters:
-                if filter_(read, matches):
+                if filter_(read, info):
                     break
         return (n, total_bp, None)
 
+    def _open_writer(
+        self,
+        file: BinaryIO,
+        file2: Optional[BinaryIO] = None,
+        force_fasta: Optional[bool] = None,
+    ):
+        assert file2 is None
+        assert not isinstance(file, (str, bytes, Path))
+        return dnaio.open(
+            file, mode="w", qualities=self.uses_qualities, fileformat="fasta" if force_fasta else None)
+
     def _filter_wrapper(self):
         return Redirector
 
@@ -322,12 +329,13 @@ class PairedEndPipeline(Pipeline):
 
     def __init__(self, pair_filter_mode, file_opener: FileOpener):
         super().__init__(file_opener)
+        self._modifiers = []  # type: List[PairedModifier]
         self._pair_filter_mode = pair_filter_mode
         self._reader = None
         # Whether to ignore pair_filter mode for discard-untrimmed filter
         self.override_untrimmed_pair_filter = False
 
-    def add(self, modifier1: Optional[Modifier], modifier2: Optional[Modifier]):
+    def add(self, modifier1: Optional[SingleEndModifier], modifier2: Optional[SingleEndModifier]):
         """
         Add a modifier for R1 and R2. One of them can be None, in which case the modifier
         will only be added for the respective read.
@@ -336,7 +344,7 @@ class PairedEndPipeline(Pipeline):
             raise ValueError("Not both modifiers can be None")
         self._modifiers.append(PairedModifierWrapper(modifier1, modifier2))
 
-    def add_both(self, modifier: Modifier):
+    def add_both(self, modifier: SingleEndModifier):
         """
         Add one modifier for both R1 and R2
         """
@@ -356,18 +364,37 @@ class PairedEndPipeline(Pipeline):
             n += 1
             if n % 10000 == 0 and progress:
                 progress.update(n)
-            total1_bp += len(read1.sequence)
-            total2_bp += len(read2.sequence)
-            matches1 = []
-            matches2 = []
+            total1_bp += len(read1)
+            total2_bp += len(read2)
+            info1 = ModificationInfo(read1)
+            info2 = ModificationInfo(read2)
             for modifier in self._modifiers:
-                read1, read2 = modifier(read1, read2, matches1, matches2)
+                read1, read2 = modifier(read1, read2, info1, info2)
             for filter_ in self._filters:
                 # Stop writing as soon as one of the filters was successful.
-                if filter_(read1, read2, matches1, matches2):
+                if filter_(read1, read2, info1, info2):
                     break
         return (n, total1_bp, total2_bp)
 
+    def _open_writer(
+        self,
+        file: BinaryIO,
+        file2: Optional[BinaryIO] = None,
+        force_fasta: Optional[bool] = None,
+    ):
+        # file and file2 must already be file-like objects because we don’t want to
+        # take care of threads and compression levels here.
+        for f in (file, file2):
+            assert not isinstance(f, (str, bytes, Path))
+        return dnaio.open(
+            file,
+            file2=file2,
+            mode="w",
+            qualities=self.uses_qualities,
+            fileformat="fasta" if force_fasta else None,
+            interleaved=file2 is None,
+        )
+
     def _filter_wrapper(self, pair_filter_mode=None):
         if pair_filter_mode is None:
             pair_filter_mode = self._pair_filter_mode
@@ -384,9 +411,7 @@ class PairedEndPipeline(Pipeline):
             return self._filter_wrapper()
 
     def _final_filter(self, outfiles):
-        writer = self._open_writer(
-            outfiles.out, outfiles.out2, interleaved=outfiles.interleaved,
-            force_fasta=outfiles.force_fasta)
+        writer = self._open_writer(outfiles.out, outfiles.out2, force_fasta=outfiles.force_fasta)
         return PairedNoFilter(writer)
 
     def _create_demultiplexer(self, outfiles):
@@ -556,7 +581,6 @@ class WorkerProcess(Process):
             orig_outfile = getattr(self._orig_outfiles, attr)
             if orig_outfile is not None:
                 output = io.BytesIO()
-                output.name = orig_outfile.name
                 setattr(output_files, attr, output)
 
         return output_files
@@ -602,7 +626,7 @@ class PipelineRunner(ABC):
     """
     A read processing pipeline
     """
-    def __init__(self, pipeline: Pipeline, progress: Progress):
+    def __init__(self, pipeline: Pipeline, progress: Progress, *args, **kwargs):
         self._pipeline = pipeline
         self._progress = progress
 
@@ -614,6 +638,12 @@ class PipelineRunner(ABC):
     def close(self):
         pass
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args):
+        self.close()
+
 
 class ParallelPipelineRunner(PipelineRunner):
     """
