@@ -9,7 +9,7 @@ import pytest
 import subprocess
 
 from cutadapt.__main__ import main
-from utils import assert_files_equal, datapath, cutpath, redirect_stderr
+from utils import assert_files_equal, datapath, cutpath
 
 # pytest.mark.timeout will not fail even if pytest-timeout is not installed
 try:
@@ -69,19 +69,6 @@ def test_discard_untrimmed(run):
     run('-b CAAGAT --discard-untrimmed', 'discard-untrimmed.fastq', 'small.fastq')
 
 
-@pytest.mark.skip(reason='Regression since switching to dnaio')
-def test_second_header_retained(run, cores):
-    """test if sequence name after the "+" is retained"""
-    run("--cores {} -e 0.12 -b TTAGACATATCTCCGTCG".format(cores), "plus.fastq", "plus.fastq")
-
-
-@pytest.mark.skip(reason='Regression since switching to dnaio')
-def test_length_tag_second_header(run, cores):
-    """Ensure --length-tag= also modifies the second header line"""
-    run("--cores {} -a GGCTTC --length-tag=length=".format(cores),
-        'SRR2040271_1.fastq', 'SRR2040271_1.fastq')
-
-
 def test_extensiontxtgz(run):
     """automatic recognition of "_sequence.txt.gz" extension"""
     run("-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz")
@@ -343,8 +330,16 @@ def test_adapter_with_u(run):
     run("-a GCCGAACUUCUUAGACUGCCUUAAGGACGU", "illumina.fastq", "illumina.fastq.gz")
 
 
-def test_bzip2(run):
-    run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.bz2')
+def test_bzip2_input(run, cores):
+    run(["--cores", str(cores), "-a", "TTAGACATATCTCCGTCG"], "small.fastq", "small.fastq.bz2")
+
+
+@pytest.mark.parametrize("extension", ["bz2", "xz", "gz"])
+def test_compressed_output(tmp_path, cores, extension):
+    out_path = str(tmp_path / ("small.fastq." + extension))
+    params = [
+        "--cores", str(cores), "-a", "TTAGACATATCTCCGTCG", "-o", out_path, datapath("small.fastq")]
+    assert main(params) is None
 
 
 if sys.version_info[:2] >= (3, 3):
@@ -358,14 +353,12 @@ def test_xz(run):
 
 def test_no_args():
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main([])
+        main([])
 
 
 def test_two_fastqs():
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main([datapath('paired.1.fastq'), datapath('paired.2.fastq')])
+        main([datapath('paired.1.fastq'), datapath('paired.2.fastq')])
 
 
 def test_anchored_no_indels(run):
@@ -386,8 +379,7 @@ def test_anchored_no_indels_wildcard_adapt(run):
 
 def test_non_iupac_characters(run):
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main(['-a', 'ZACGT', datapath('small.fastq')])
+        main(['-a', 'ZACGT', datapath('small.fastq')])
 
 
 def test_unconditional_cut_front(run):
@@ -544,20 +536,17 @@ def test_linked_info_file(tmpdir):
 
 def test_linked_anywhere():
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main(['-b', 'AAA...TTT', datapath('linked.fasta')])
+        main(['-b', 'AAA...TTT', datapath('linked.fasta')])
 
 
 def test_anywhere_anchored_5p():
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main(['-b', '^AAA', datapath('small.fastq')])
+        main(['-b', '^AAA', datapath('small.fastq')])
 
 
 def test_anywhere_anchored_3p():
     with pytest.raises(SystemExit):
-        with redirect_stderr():
-            main(['-b', 'TTT$', datapath('small.fastq')])
+        main(['-b', 'TTT$', datapath('small.fastq')])
 
 
 def test_fasta(run):
@@ -709,15 +698,46 @@ def test_adapter_order(run):
     run("-a CCGGG -g ^AAACC", "adapterorder-ag.fasta", "adapterorder.fasta")
 
 
-@pytest.mark.skip(reason="Not implemented")
-def test_reverse_complement_not_normalized(run):
-    run("--rc=yes -g ^TTATTTGTCT -g ^TCCGCACTGG",
-        "revcomp-notnormalized-single.fastq", "revcomp.1.fastq")
-
-
 def test_reverse_complement_normalized(run):
     run(
         "--revcomp -g ^TTATTTGTCT -g ^TCCGCACTGG",
         "revcomp-single-normalize.fastq",
         "revcomp.1.fastq",
     )
+
+
+def test_reverse_complement_and_info_file(run, tmp_path, cores):
+    info_path = str(tmp_path / "info.txt")
+    run(
+        [
+            "--revcomp",
+            "-g",
+            "^TTATTTGTCT",
+            "-g",
+            "^TCCGCACTGG",
+            "--info-file",
+            info_path,
+        ],
+        "revcomp-single-normalize.fastq",
+        "revcomp.1.fastq",
+    )
+    with open(info_path) as f:
+        lines = f.readlines()
+    assert len(lines) == 6
+    assert lines[0].split("\t")[0] == "read1/1"
+    assert lines[1].split("\t")[0] == "read2/1 rc"
+
+
+def test_max_expected_errors(run, cores):
+    run("--max-ee=0.9", "maxee.fastq", "maxee.fastq")
+
+
+def test_max_expected_errors_fasta(tmp_path):
+    path = tmp_path / "input.fasta"
+    path.write_text(">read\nACGTACGT\n")
+    main(["--max-ee=0.001", "-o", "/dev/null", str(path)])
+
+
+def test_warn_if_en_dashes_used():
+    with pytest.raises(SystemExit):
+        main(["–q", "25", "-o", "/dev/null", "in.fastq"])
