@@ -5,7 +5,7 @@ from itertools import product
 import pytest
 
 from cutadapt.__main__ import main
-from utils import assert_files_equal, datapath, cutpath, redirect_stderr
+from utils import assert_files_equal, datapath, cutpath
 
 
 @pytest.fixture
@@ -131,9 +131,8 @@ def test_no_trimming():
 
 
 def test_missing_file(tmpdir):
-    with redirect_stderr():
-        with pytest.raises(SystemExit):
-            main(["--paired-output", str(tmpdir.join("out.fastq")), datapath("paired.1.fastq")])
+    with pytest.raises(SystemExit):
+        main(["--paired-output", str(tmpdir.join("out.fastq")), datapath("paired.1.fastq")])
 
 
 def test_first_too_short(tmpdir, cores):
@@ -144,14 +143,13 @@ def test_first_too_short(tmpdir, cores):
         lines = lines[:-4]
     trunc1.write("".join(lines))
 
-    with redirect_stderr():
-        with pytest.raises(SystemExit):
-            main([
-                "-o", "/dev/null",
-                "--paired-output", str(tmpdir.join("out.fastq")),
-                "--cores", str(cores),
-                str(trunc1), datapath("paired.2.fastq")
-            ])
+    with pytest.raises(SystemExit):
+        main([
+            "-o", "/dev/null",
+            "--paired-output", str(tmpdir.join("out.fastq")),
+            "--cores", str(cores),
+            str(trunc1), datapath("paired.2.fastq")
+        ])
 
 
 def test_second_too_short(tmpdir, cores):
@@ -162,14 +160,13 @@ def test_second_too_short(tmpdir, cores):
         lines = lines[:-4]
     trunc2.write("".join(lines))
 
-    with redirect_stderr():
-        with pytest.raises(SystemExit):
-            main([
-                "-o", "/dev/null",
-                "--paired-output", str(tmpdir.join("out.fastq")),
-                "--cores", str(cores),
-                datapath("paired.1.fastq"), str(trunc2)
-            ])
+    with pytest.raises(SystemExit):
+        main([
+            "-o", "/dev/null",
+            "--paired-output", str(tmpdir.join("out.fastq")),
+            "--cores", str(cores),
+            datapath("paired.1.fastq"), str(trunc2)
+        ])
 
 
 def test_unmatched_read_names(tmpdir, cores):
@@ -327,10 +324,24 @@ def test_interleaved_neither_nor(tmpdir):
     p1 = str(tmpdir.join("temp-paired.1.fastq"))
     p2 = str(tmpdir.join("temp-paired.2.fastq"))
     params = "-a XX --interleaved".split()
-    with redirect_stderr():
-        params += ["-o", p1, "-p1", p2, "paired.1.fastq", "paired.2.fastq"]
-        with pytest.raises(SystemExit):
-            main(params)
+    params += ["-o", p1, "-p1", p2, "paired.1.fastq", "paired.2.fastq"]
+    with pytest.raises(SystemExit):
+        main(params)
+
+
+def test_interleaved_untrimmed_output(tmpdir):
+    o1 = str(tmpdir.join("out.1.fastq"))
+    o2 = str(tmpdir.join("out.2.fastq"))
+    untrimmed = str(tmpdir.join("untrimmed.interleaved.fastq"))
+    main([
+        "--interleaved",
+        "-a", "XXXX",
+        "-o", o1,
+        "-p", o2,
+        "--untrimmed-output", untrimmed,
+        datapath("interleaved.fastq")
+    ])
+    assert_files_equal(datapath("interleaved.fastq"), untrimmed)
 
 
 def test_pair_filter_both(run_paired, cores):
@@ -504,25 +515,40 @@ def test_pair_adapters_demultiplexing(tmpdir):
         assert_files_equal(cutpath(name), str(tmpdir.join(name)))
 
 
-def test_combinatorial_demultiplexing(tmpdir):
+@pytest.mark.parametrize("discarduntrimmed", (False, True))
+def test_combinatorial_demultiplexing(tmpdir, discarduntrimmed):
     params = "-g A=^AAAAAAAAAA -g C=^CCCCCCCCCC -G G=^GGGGGGGGGG -G T=^TTTTTTTTTT".split()
     params += ["-o", str(tmpdir.join("combinatorial.{name1}_{name2}.1.fasta"))]
     params += ["-p", str(tmpdir.join("combinatorial.{name1}_{name2}.2.fasta"))]
     params += [datapath("combinatorial.1.fasta"), datapath("combinatorial.2.fasta")]
+    combinations = [
+        # third column says whether the file must exist
+        ("A", "G", True),
+        ("A", "T", True),
+        ("C", "G", True),
+        ("C", "T", True),
+    ]
+    if discarduntrimmed:
+        combinations += [
+            ("unknown", "G", False),
+            ("A", "unknown", False),
+        ]
+        params += ["--discard-untrimmed"]
+    else:
+        combinations += [
+            ("unknown", "G", True),
+            ("A", "unknown", True),
+        ]
     assert main(params) is None
-    for (name1, name2) in [
-        ("A", "G"),
-        ("A", "T"),
-        ("C", "G"),
-        ("C", "T"),
-        ("unknown", "G"),
-        ("A", "unknown"),
-    ]:
+    for (name1, name2, should_exist) in combinations:
         for i in (1, 2):
             name = "combinatorial.{name1}_{name2}.{i}.fasta".format(name1=name1, name2=name2, i=i)
             path = cutpath(os.path.join("combinatorial", name))
-            assert tmpdir.join(name).check(), "Output file missing"
-            assert_files_equal(path, str(tmpdir.join(name)))
+            if should_exist:
+                assert tmpdir.join(name).check(), "Output file missing"
+                assert_files_equal(path, str(tmpdir.join(name)))
+            else:
+                assert not tmpdir.join(name).check(), "Output file should not exist"
 
 
 def test_info_file(tmpdir):
